'use strict';

// https://pannellum.org/documentation/overview/

var data_version = 2;
var num_chapters = 5;

var state = null;
var randseed = 0;
var inspecting = false;
var revealing = false;
var using_localstorage = false;

var viewer = null;

function reset_state()
{
    state = {
        dataversion: data_version,
        titleshown: false,
        invopen: true,
        spoilerpref: false,
        read: [ false, false, false, false, false ],
        gotlens: false,
        foundorder: [],
        seen: {}, // book -> item
        found: {}, // item -> bool
        searched: {}, // book -> bool
        used: {}, // item -> book
        filled: {} // book -> item
    };
    
    save_state();
}

function save_state()
{
    if (using_localstorage) {
        localStorage.setItem('booksphere_state', JSON.stringify(state));
    }
}

function remove_overlay()
{
    $('.Overlay').remove();
}

var title_html = "<h1>Tour Bibliotekh</h1>\n\
<h2>An interactive bucharesque</h2>\n\
<h2 class=\"Subheader\">Copyright 2020 by Andrew Plotkin \u2014 release 5</h2>\n\
<p>Created for the Italo Calvino game jam,\n\
\u201C\<a href=\"https://itch.io/jam/calvino2020\" target=\"_blank\">If on a summer\u2019s night a game designer...</a>\u201D</p>\n\
<p>For the designer\u2019s notes, <a href=\"about.html\" target=\"_blank\">see here</a>.</p>\n\
<p class=\"SpoilerPrefPara\">\n\
Book-spoiler mode:\n\
<a id=\"spoilerpref\" href=\"#\" onclick=\"return evhan_togglespoiler();\">???</a>\n\
<br>\n\
(Titles and authors of books <span id=\"spoilerblurb\">???</span>)</p>\n\
<p class=\"ResetPara\"><span id=\"resetwarning\">Are you sure?<br>Every item you have found will be returned to its original book.<br></span><a id=\"resetquery\" href=\"#\" onclick=\"return evhan_resetquery();\">(Reset all progress)</a></p>\n\
";

function display_title(key)
{
    $('.Overlay').remove();

    var item = itemmap[key];
    
    var overel = $('<div>', { class:'Overlay' });
    overel.on('click', remove_overlay);

    var coloel = $('<div>', { class:'TextColOut NarrowCol' });
    overel.append(coloel);
    
    var colel = $('<div>', { class:'TextCol NarrowCol' });
    coloel.append(colel);

    var screl = $('<div>', { class:'ScrollCol' });
    screl.on('click', function(ev) {
        // permit links to bubble, *except* the reset link (sigh)
        if (ev.target && ev.target.nodeName == "A" &&
            (ev.target.id != 'resetquery' && ev.target.id != 'spoilerpref'))
            return true;
        return false;
    });
    colel.append(screl);
    
    var el = $('<div>');
    el.html(title_html);
    screl.append(el);

    $('#panbox').append(overel);
    
    adjust_spoilerpref();
}

function display_maintext(chapter)
{
    $('.Overlay').remove();

    state.read[chapter] = true;
    adjust_read_status();
    
    var overel = $('<div>', { class:'Overlay' });
    overel.on('click', remove_overlay);

    var coloel = $('<div>', { class:'TextColOut WideCol' });
    overel.append(coloel);
    
    var colel = $('<div>', { class:'TextCol WideCol' });
    coloel.append(colel);

    var screl = $('<div>', { class:'ScrollCol' });
    screl.on('click', function() { return false; });
    colel.append(screl);

    for (var ix=0; ix<num_chapters; ix++) {
        var textdiv;
        if (!state.read[ix]) {
            textdiv = $('<p>').text('...');
        }
        else {
            textdiv = $('<div>');
            if (ix == chapter) {
                textdiv.addClass('CurrentPara');
            }
            else {
                textdiv.addClass('NonCurrentPara');
            }
            textdiv.html(chaptertext[ix]);
        }
        screl.append(textdiv);
        screl.append($('<hr>'));
    }

    $('#panbox').append(overel);

    var curpara = $('.CurrentPara');
    if (curpara.length) {
        var val = curpara.position().top;
        val = val-30;
        if (val > 0)
            screl.scrollTop(val);
    }
    
    save_state();
}

function display_book(key)
{
    $('.Overlay').remove();

    var book = bookmap[key];
    inspecting = false;
    revealing = false;
    randseed = Math.floor(Math.random() * 0xFFFF);
    
    var overel = $('<div>', { class:'Overlay' });
    overel.on('click', remove_overlay);

    var coloel = $('<div>', { class:'TextColOut NarrowCol' });
    overel.append(coloel);
    
    var colel = $('<div>', { class:'TextCol NarrowCol' });
    coloel.append(colel);

    var screl = $('<div>', { class:'ScrollCol' });
    screl.on('click', function() { return false; });
    colel.append(screl);

    build_book(key, screl);

    $('#panbox').append(overel);

    if (!state.seen[key]) {
        state.seen[key] = true;
        adjust_bookspot(key);
        save_state();
    }
}

function display_invitem(key)
{
    $('.Overlay').remove();

    var item = itemmap[key];
    
    var overel = $('<div>', { class:'Overlay' });
    overel.on('click', remove_overlay);

    var coloel = $('<div>', { class:'TextColOut NarrowCol' });
    overel.append(coloel);
    
    var colel = $('<div>', { class:'TextCol NarrowCol' });
    coloel.append(colel);

    var screl = $('<div>', { class:'ScrollCol' });
    screl.on('click', function() { return false; });
    colel.append(screl);

    var el = $('<p>');
    el.text('You have acquired ' + item.name + '.');
    screl.append(el);

    var textdiv = $('<div>', { class:'InvDescription' });
    textdiv.html(item.html);
    screl.append(textdiv);

    $('#panbox').append(overel);
}

function build_book(key, screl)
{
    var book = bookmap[key];
    var itemkey = book.item;
    var item = null;
    if (itemkey)
        item = itemmap[itemkey];
    
    if (screl == null) {
        screl = $('.ScrollCol');
        screl.empty();
    }

    if (state.spoilerpref) {
        if (!revealing) {
            var el = $('<p>', { class:'ActionLink' });
            var ael = $('<a>', { href:'#' });
            ael.text('(Reveal title and author)');
            ael.on('click', { key:key }, evhan_revealspoiler);
            el.append(ael);
            screl.append(el);
        }
        else {
            var el = $('<p>');
            el.append($('<span>').text('('));
            el.append($('<em>').text(book.title));
            el.append($('<span>').text(', '));
            el.append($('<span>').text(book.author));
            el.append($('<span>').text(')'));
            screl.append(el);
        }
    }
    
    if (state.filled[key]) {
        var hasitem = state.filled[key];
        var el = $('<p>', { class:'ActionLink' });
        el.text(random_choice(say_item_emplaced).replace('$$', itemmap[hasitem].name));
        screl.append(el);
        var snip = book.snippets[hasitem];
        var textdiv = $('<div>', { class:'BookText' });
        textdiv.html(snip.html);
        screl.append(textdiv);
    }
    else {
        var textdiv = $('<div>', { class:'BookText' });
        textdiv.html(book.html);
        screl.append(textdiv);

        var items = current_inventory();

        for (var itkey of items) {
            if (book.snippets[itkey]) {
                var el = $('<p>', { class:'ActionLink' });
                el.text('You could place ');
                var ael = $('<a>', { href:'#' });
                ael.text(itemmap[itkey].name);
                ael.on('click', { key:key, itemkey:itkey }, evhan_insertitem);
                el.append(ael);
                el.append($('<span>').text(' into this book.'));
                screl.append(el);
            }
        }
    }
    
    if (book.containslens) {
        if (!state.gotlens) {
            if (!inspecting) {
                var el = $('<p>', { class:'ActionLink' });
                el.text('You notice a raised shadow on the endpaper, as if something were ');
                var ael = $('<a>', { href:'#' });
                ael.text('concealed in the back cover.');
                ael.on('click', { key:key }, evhan_inspectbook);
                el.append(ael);
                screl.append(el);
            }
            else {
                var el = $('<p>', { class:'ActionLink' });
                el.text('You peel back the endpaper, uncovering a shallow hollow in the back cover. ');
                var ael = $('<a>', { href:'#' });
                ael.text('A compact hand lens');
                ael.on('click', { key:key }, evhan_takeitem);
                el.append(ael);
                el.append($('<span>').text(' rests within.'));
                screl.append(el);
            }
        }
    }

    if (state.gotlens) {
        if (!state.searched[key]) {
            if (!inspecting) {
                var el = $('<p>', { class:'ActionLink' });
                var ael = $('<a>', { href:'#' });
                ael.text(random_choice(say_inspect_book));
                ael.on('click', { key:key }, evhan_inspectbook);
                el.append(ael);
                screl.append(el);
            }
            else if (!itemkey) {
                // should not happen
            }
            else {
                var el = $('<p>', { class:'ActionLink' });
                el.text('You peer carefully through the lens and discover ');
                var ael = $('<a>', { href:'#' });
                ael.text(item.name + '.');
                ael.on('click', { key:key }, evhan_takeitem);
                el.append(ael);
                screl.append(el);
            }
        }
        else {
            if (inspecting) {
                if (book.containslens) {
                    var el = $('<p>', { class:'ActionLink' });
                    el.text('You take the hand lens.');
                    screl.append(el);
                }
                else if (itemkey) {
                    // should not happen
                }
                else {
                    var el = $('<p>', { class:'ActionLink' });
                    el.text(random_choice(say_nothing_within));
                    screl.append(el);
                }
            }
            else {
                var el = $('<p>', { class:'ActionLink' });
                el.text(random_choice(say_already_inspected));
                screl.append(el);
            }
        }
    }

}

function build_inventory()
{
    var bodyel = $('#invbody');
    if (!state.gotlens) {
        return;
    }
    
    bodyel.empty();

    var items = current_inventory();

    if (!state.invopen) {
        $('#invtoggle').text('(show)');
    }
    else {
        $('#invtoggle').text('(hide)');
    }

    if (!state.invopen) {
        var val = 1 + items.length;
        var ael = $('<a>', { href:'#'} );
        if (val == 1)
            ael.text('One item');
        else
            ael.text(val + ' items');
        ael.on('click', evhan_invtoggle);
        bodyel.append($('<div>').append(ael));
    }
    else if (!items.length) {
        bodyel.append($('<div>').text('A hand lens'));
    }
    else {
        bodyel.append($('<div>').text('A hand lens, with which you have discovered:'));
        var el = $('<div>');

        for (var ix=0; ix<items.length; ix++) {
            if (ix == 0) {
                var ael = $('<span>').text('...');
                el.append(ael);
            }
            else if (ix == 1 && items.length == 2) {
                var ael = $('<span>').text(' and ');
                el.append(ael);
            }
            else if (ix+1 == items.length) {
                var ael = $('<span>').text(', and ');
                el.append(ael);
            }
            else {
                var ael = $('<span>').text(', ');
                el.append(ael);
            }
            
            var key = items[ix];
            var item = itemmap[key];

            var ael = $('<a>', { href:'#' }).text(item.name);
            ael.on('click', { key:key }, evhan_invclick);
            el.append(ael);
        }

        el.append($('<span>').text('.'));
        
        bodyel.append(el);
    }
}

function adjust_read_status()
{
    var el = $('#panbox');
    
    for (var ix=0; ix<num_chapters; ix++) {
        if (!state.read[ix]) {
            el.removeClass('Read_TextSpot_'+ix);
        }
        else {
            el.addClass('Read_TextSpot_'+ix);
        }
    }
}

function adjust_spoilerpref()
{
    if (state.spoilerpref) {
        $('#spoilerpref').text('YES');
        $('#spoilerblurb').text('can be revealed when examining a book.');
    }
    else {
        $('#spoilerpref').text('\xA0NO');
        $('#spoilerblurb').text('are not shown. Can you figure them out?');
    }
}

function adjust_bookspot(key)
{
    var book = bookmap[key];
    if (!book)
        return;

    var searched = state.searched[key];
    var filled = (book.snippetcount == 0 || state.filled[key]);
    
    var color = 'fresh';
    if (state.seen[key])
        color = 'seen';
    if (searched || filled)
        color = 'touched';
    if (searched && filled)
        color = 'done';

    if (book.containslens && !state.gotlens)
        color = 'fresh';
    
    var el = $('.book_'+key);
    if (el.length) {
        el.get(0).dataset.color = color;
    }
}

function current_inventory()
{
    var items = [];

    for (var key of state.foundorder) {
        if (state.found[key] && !state.used[key]) {
            items.push(key);
        }
    }
    
    return items;
}

function evhan_infoclick(ev, args)
{
    if (args.type == 'book')
        display_book(args.key);
    else if (args.type == 'text')
        display_maintext(args.index);

    return false;
}

function evhan_invclick(ev)
{
    var args = ev.data;
    var key = args.key;
    
    display_invitem(key);

    return false;
}

function evhan_inspectbook(ev)
{
    var args = ev.data;
    var key = args.key;

    var book = bookmap[key];

    var itemkey = book.item;
    if (!itemkey) {
        state.searched[key] = true;
        save_state();
    }

    inspecting = true;
    build_book(key);
    adjust_bookspot(key);

    return false;
}

function evhan_revealspoiler(ev)
{
    var args = ev.data;
    var key = args.key;

    revealing = true;
    build_book(key);

    return false;
}

function evhan_takeitem(ev)
{
    var args = ev.data;
    var key = args.key;
    
    var book = bookmap[key];
    var itemkey = book.item;
    
    if (book.containslens) {
        state.gotlens = true;
        $('#panbox').addClass('InventoryActive');
        build_inventory();
        build_book(key);
        adjust_bookspot(key);
    }
    else if (itemkey) {
        acquire(itemkey);
        build_inventory();
        display_invitem(itemkey);
        adjust_bookspot(key);
    }

    save_state();

    return false;
}

function evhan_insertitem(ev)
{
    var args = ev.data;
    var key = args.key;
    var itemkey = args.itemkey;

    state.used[itemkey] = key;
    state.filled[key] = itemkey;

    build_book(key);
    build_inventory();
    adjust_bookspot(key);

    save_state();

    return false;
}

function evhan_showabout(ev)
{
    display_title();
    return false;
}

function evhan_resetquery(ev)
{
    if (!$('.ResetPara').hasClass('ResetWarning')) {
        $('.ResetPara').addClass('ResetWarning');
    }
    else {
        reset_state();
        document.location.reload();
    }
    
    return false;
}

function evhan_togglespoiler(ev)
{
    state.spoilerpref = !(state.spoilerpref);
    adjust_spoilerpref();
    save_state();
    return false;
}

function evhan_strangeloop(ev)
{
    var el = $('.StrangeLoopText');
    if (!el.length) {
        el = $('<p>', { class:'StrangeLoopText' }).text('You\u2019ve been on this merry-go-round before, haven\u2019t you?');
        $('.BookText').append(el);
    }
    
    return false;
}

function evhan_invtoggle(ev)
{
    state.invopen = !(state.invopen);
    build_inventory();
    save_state();
    return false;
}

function evhan_sceneload(ev)
{
    var scene = scenelist[viewer.getScene()];
    for (var spot of scene.hotSpots) {
        var args = spot.clickHandlerArgs;
        if (!(args && args.type == 'book'))
            continue;
        adjust_bookspot(args.key);
    }
}

var last_down_event = null;

function evhan_panmousedown(ev)
{
    last_down_event = ev;
}

function evhan_panmouseup(ev)
{
    if (last_down_event != null) {
        var panbox = $('#panbox');
        var dur = ev.timeStamp - last_down_event.timeStamp;
        var dist = Math.hypot(ev.x - last_down_event.x, ev.y - last_down_event.y) / panbox.width();
        if (dur < 1000 && dist < 0.1) {
            if (dur < 500 || dist < 0.05) {
                if (!panbox.hasClass('Flash')) {
                    panbox.addClass('Flash');
                    window.setTimeout(function() {
                        panbox.removeClass('Flash');
                    }, 200);
                }
            }
        }
    }
    last_down_event = null;
}

function acquire(key)
{
    if (state.found[key])
        return;

    var item = itemmap[key];
    
    state.found[key] = true;
    state.foundorder.push(key);
    state.searched[item.book] = true;

    save_state();
}

var say_item_emplaced = [
    '(...$$ is emplaced within.)',
    '(You have placed $$ within.)',
    '(You have inserted $$.)',
    '(You have put $$ here.)',
    '(The book now contains $$.)'
];

var say_inspect_book = [
    'Inspect the book with the lens.',
    'Examine the book with the lens.',
    'Peer at the book with the lens.'
];

var say_nothing_within = [
    'You discover nothing within.',
    'You find nothing within.',
    'You discover nothing inside.',
    'You find nothing inside.'
];

var say_already_inspected = [
    '(You have inspected this book already.)',
    '(You have already inspected this book.)',
    '(You have examined this book already.)',
    '(You have already examined this book.)'
];

function random_choice(ls)
{
    var count = ls.length;
    if (count == 0)
        return null;
    if (count == 1)
        return ls[0];
    
    var val = (randseed % count);
    if (val < 0 || val >= count)
        return ls[0];
    return ls[val];
}

function purloin()
{
    for (var ix=0; ix<num_chapters; ix++) {
        state.read[ix] = true;
    }
    adjust_read_status();

    state.gotlens = true;
    $('#panbox').addClass('InventoryActive');

    build_inventory();
    
    save_state();
}

function init()
{
    try {
        localStorage.setItem('booksphere_test', 'test');
        var val = localStorage.getItem('booksphere_state');
        if (val == null) {
            reset_state();
        }
        else {
            state = JSON.parse(val);
            if ((!state) || state.dataversion != data_version) {
                state = null;
                reset_state();
            }
        }
        using_localstorage = true;
    }
    catch (ex) {
        console.log('Unable to use localStorage; state will not be saved');
        reset_state();
    }

    var debug = $('body').hasClass('Debug');
    
    viewer = pannellum.viewer('pandiv', {
        default: {
            firstScene: 'start',
            type: 'equirectangular',
            autoLoad: true,
            friction: 0.2,
            maxPitch: 110, minPitch: -110,
            sceneFadeDuration: 700,
            compass: false,
            hotSpotDebug: debug
        },

        scenes: scenelist,

        showFullscreenCtrl: false
    });
    viewer.on('load', evhan_sceneload);
    viewer.on('mousedown', evhan_panmousedown);
    viewer.on('mouseup', evhan_panmouseup);

    var aboutpane = $('.pnlm-about-msg');
    aboutpane.html('<a href="about.html" target="_blank">Game Designer\'s Notes</a>');

    if (state.gotlens)
        $('#panbox').addClass('InventoryActive');
    
    adjust_read_status();
    build_inventory();

    if (!state.titleshown) {
        state.titleshown = true;
        display_title();
    }

    save_state();
}
